import numpy as np
import arithmetics
from scipy import optimize
np.set_printoptions(precision=4, suppress=True)
np.set_printoptions(linewidth=180)


'''
def ep_ode(A, B, T, x0, y0, dx_dt0=None, dy_dt0=None, damp=1/16, mass=1, dt=0.001, smoothmax_corner=0.01, eps=1e-6):
    t = 0
    x = x0
    y = y0
    dx_dt = dx_dt0
    dy_dt = dy_dt0
    ep = np.inf
    if dx_dt0 is None:
        dx_dt = np.zeros_like(x0)
    if dy_dt0 is None:
        dy_dt = np.zeros_like(y0)
    while t < T:
        u = A @ y
        v = B.T @ x
        ep = u.max(axis=0) - x.T @ u + v.max(axis=0) - y.T @ v
        dep_dx = B @ softmax(v, corner=smoothmax_corner) - u
        dep_dy = A.T @ softmax(u, corner=smoothmax_corner) - v
        
        dx = dx_dt * dt
        dy = dy_dt * dt
        ddx_dt = dep_dx
        ddy_dt = dep_dy
        
        
        
        t += dt
        x += dx
        y += dy
        dx_dt += ddx_dt
        dy_dt += ddy_dt
    return x, y, dx_dt, dy_dt, ep
'''

def ep_downhill(A, B, x0, y0, T, smoothmax_corner=0.01, dt=0.001, eps=1e-6):
    x = x0
    y = y0
    ep = np.inf
    for t in range(T):
        epA = arithmetics.smoothmax(A @ y, corner=smoothmax_corner) - x.T @ A @ y
        epB = arithmetics.smoothmax(B.T @ x, corner=smoothmax_corner) - x.T @ B @ y
        ep = epA + epB
        Dep_Dx = B @ arithmetics.softmax(B.T @ x, corner=smoothmax_corner) - (A+B) @ y
        Dep_Dy = A.T @ arithmetics.softmax(A @ y, corner=smoothmax_corner) - (A.T + B.T) @ x

        dx = -Dep_Dx
        N = [np.ones_like(x)]
        for i in range(len(x)):
            if x[i] <= eps and dx[i] < -eps:
                N.append(np.zeros_like(x))
                N[-1][i] = -1
        N = np.concatenate(N, axis=1)
        coef = np.linalg.pinv(N.T @ N) @ N.T @ dx
        dx -= N @ coef

        dy = -Dep_Dy
        N = [np.ones_like(y)]
        for j in range(len(y)):
            if y[j] <= eps and dy[j] < -eps:
                N.append(np.zeros_like(y))
                N[-1][j] = -1
        N = np.concatenate(N, axis=1)
        coef = np.linalg.pinv(N.T @ N) @ N.T @ dy
        dy -= N @ coef

        dt_x = x / (-dx)
        dt_x[dx >= -eps] = np.inf
        dt_x = dt_x.min(axis=0)[0]
        dt_x = min(dt_x, dt)

        dt_y = y / (-dy)
        dt_y[dy >= -eps] = np.inf
        dt_y = dt_y.min(axis=0)[0]
        dt_y = min(dt_y, dt)

        x += dx * dt_x
        y += dy * dt_y
    return x, y, ep


def test():
    A = np.array([[13800, -36365, -18522],
                  [36700, 13800, -36200],
                  [18522, 36365, 13800]]) / 10000
    B = np.array([[29500, 10383, -33139],
                  [21600, 29500, 70700],
                  [33139, -10383, 29500]]) / 10000
    print("A:\n", A)
    print("B:\n", B)

    # ''' Goofspiel
    x0 = arithmetics.rand_f(1, 2).T
    y0 = arithmetics.rand_f(1, 2).T
    x1 = np.array([[0.0000, 0.0690, 0.9310]]).T
    y1 = np.array([[0.7334, 0.0000, 0.2666]]).T
    # '''
    x = x0
    y = y0
    ep = np.inf
    for t in range(16):
        print("t:", t, end=", ")
        print("ep:", ep, end=", ")
        print("x:", x.T, end=", ")
        print("y:", y.T)
        x, y, ep = ep_downhill(A, B, x, y, 16, smoothmax_corner=0.01, dt=0.001)

    print("ep_downhill:")
    print("x:", x.T)
    print("y:", y.T)
    print("")


def relu(x, eps=1e-2):
    return x
    # return (np.sqrt(x**2 + eps**2) + x) / 2


def step(x, eps=1e-2):
    return x * 0 + 1
    # return (x / np.sqrt(x**2 + eps**2) + 1) / 2


def imp(x, eps=1e-2):
    return x * 0
    # return eps**2 / np.sqrt(x**2 + eps**2)**3


if __name__ == "__main__":
    # test()
    A = np.array([[13800, -36365, -18522],
                  [36700, 13800, -36200],
                  [18522, 36365, 13800]]) / 10000
    B = np.array([[29500, 10383, -33139],
                  [21600, 29500, 70700],
                  [33139, -10383, 29500]]) / 10000

    A = np.array([[2, -1], [-1, 1]]) / 1.0
    B = np.array([[1, -1], [-1, 2]]) / 1.0
    m, n = A.shape
    print("A:\n", A)
    print("B:\n", B)
    print("A+B:\n", A + B)

    x = arithmetics.rand_f(1, m-1).T
    y = arithmetics.rand_f(1, n-1).T
    x1 = np.array([[0.0000, 0.0690, 0.9310]]).T
    y1 = np.array([[0.7334, 0.0000, 0.2666]]).T
    a = np.array([[0.0]])
    b = np.array([[0.0]])
    a1 = x.T @ A @ y
    b1 = x.T @ B @ y
    v = np.random.randn(2 * m + 2 * n + 2, 1)
    v1 = np.array([[-2.4156, -0.0003, 0.0000,
                   +0.0000, -7.9948, 0.0000,
                   -0.0000, -0.1189, -1.8847,
                   -1.4664,  0.0000, -0.5336,
                   -1.7263, -3.2343]]).T

    for t in range(16):
        print(t, "-" * 64)
        print("[x; y; a; b].T:")
        print(np.concatenate([x, y, a, b], axis=0).T)
        print("v.T:")
        print(v.T)

        f = a + b - x.T @ (A + B) @ y
        H = np.concatenate([relu(-x),
                            relu(-y),
                            relu(A @ y - a),
                            relu(B.T @ x - b),
                            np.ones([1, m]) @ x - 1,
                            np.ones([1, n]) @ y - 1], axis=0)
        L = f + v.T @ H
        print("f.T")
        print(f)
        print("H.T")
        print(H.T)
        print("L:")
        print(L)

        Df = np.concatenate([-(A + B) @ y, -(A + B).T @ x, np.array([[1]]), np.array([[1]])], axis=0)
        DH = np.zeros([m + n + 2, 2*m + 2*n + 2])

        DHxx = np.diag(-step(-x[:, 0]))
        DHxv = B @ np.diag(step(B.T @ x - b)[:, 0])
        DHxX = np.ones([m, 1])
        DHyy = np.diag(-step(-y[:, 0]))
        DHyu = A.T @ np.diag(step(A @ y - a)[:, 0])
        DHyY = np.ones([n, 1])
        DHau = -step(A @ y - a).T
        DHbv = -step(B.T @ x - b).T

        DH[0:m, 0:m] = DHxx
        DH[0:m, 2*m+n:2*m+2*n] = DHxv
        DH[0:m, 2*m+2*n] = DHxX[:, 0]
        DH[m:m+n, m:m+n] = DHyy
        DH[m:m+n, m+n:2*m+n] = DHyu
        DH[m:m+n, 2*m+2*n+1] = DHyY[:, 0]
        DH[m+n, m+n:2*m+n] = DHau
        DH[m+n+1, 2*m+n:2*m+2*n] = DHbv

        DL = np.concatenate([Df + DH @ v, H], axis=0)
        print("Df.T")
        print(Df.T)
        print("DH")
        print(DH)
        print("DL.T:")
        print(DL.T)

        D2f = np.zeros([m+n+2, m+n+2])
        D2f[0:m, m:m+n] = -(A + B)
        D2f[m:m+n, 0:m] = -(A + B).T

        D2L = np.zeros([3*m+3*n+4, 3*m+3*n+4])
        D2L[0:m+n+2, 0:m+n+2] = D2f
        for k in range(m):
            D2L[k, k] += imp(-x[k, 0]) * v[k, 0]
        for k in range(n):
            D2L[m+k, m+k] += imp(-y[k, 0]) * v[m+k, 0]
        for k in range(m):
            ak = A[k:k+1, :].T
            wk = (imp(ak.T@y - a) * v[m+n+k, 0])[0, 0]
            D2L[m:m+n, m:m+n] += ak @ ak.T * wk
            D2L[m:m+n, m+n] -= ak[:, 0] * wk
            D2L[m+n, m:m+n] -= ak[:, 0] * wk
            D2L[m+n, m+n] += wk
        for k in range(n):
            bk = B[:, k:k+1]
            wk = (imp(bk.T@x - b) * v[2*m+n+k, 0])[0, 0]
            D2L[0:m, 0:m] += bk @ bk.T * wk
            D2L[0:m, m+n+1] -= bk[:, 0] * wk
            D2L[m+n+1, 0:m] -= bk[:, 0] * wk
            D2L[m+n+1, m+n+1] += wk
        D2L[0:m+n+2, m+n+2:3*m+3*n+4] = DH
        D2L[m+n+2:3*m+3*n+4, 0:m+n+2] = DH.T
        # print("D2f")
        # print(D2f)
        # print("D2L")
        # print(D2L)

        # gradient descent
        E = DL.T @ DL
        delta = -2 * D2L @ DL * 0.001
        print("E.T")
        print(E.T)
        print("delta.T")
        print(delta.T)
        x += delta[0:m, :]
        y += delta[m:m+n, :]
        a += delta[m+n, 0]
        b += delta[m+n+1, 0]
        v += delta[m+n+2:3*m+3*n+4, :]
        v = DH.T @ np.linalg.pinv(DH @ DH.T) @ Df
        print(v.T)
